import os

from flask import Flask

import yabc.server.sql_backend
import yabc.server.yabc_api


def create_app(test_config=None):
    """
    Generate the flask app. Which database to use, highest priority first:

    1) any DATABASE value set in test_config 
    2) any DATABASE value set in config.py
    3) sqlite:///{app.instance_path}/yabc.sqlite

    """

    app = Flask(__name__, instance_relative_config=True)
    app.config.from_mapping(
        SECRET_KEY="dev",
        DATABASE="sqlite:///{}".format(os.path.join(app.instance_path, "yabc.sqlite")),
    )

    if test_config is None:
        app.config.from_pyfile("config.py", silent=True)
    else:
        app.config.from_mapping(test_config)
    try:
        os.makedirs(app.instance_path)
    except OSError:
        pass

    app.register_blueprint(yabc.server.yabc_api.bp)
    yabc.server.sql_backend.init_app(app)

    return app



"""
Calculate the cost basis.
"""
import csv
import decimal
import io
from decimal import Decimal
from typing import Callable
from typing import Sequence

from yabc import coinpool
from yabc import ohlcprovider
from yabc import transaction
from yabc.costbasisreport import CostBasisReport
from yabc.transaction import Transaction
from yabc.transaction import is_fiat

__author__ = "Robert Karl <robertkarljr@gmail.com>"


def _split_coin_to_add(coin_to_split, amount, trans):
    # type:  (transaction.Transaction, Decimal, transaction.Transaction) -> transaction.Transaction
    """
    Create a coin to be added back to the pool.

    TODO: For creating an audit trail, we should track the origin of the split
      coin, ie. was it generated from mining, a gift, or purchased?  This could
      be similar to the way we track the origin coin in CBRs.

    :param coin_to_split: a Transaction, either a BUY or an TRADE_INPUT
    :param amount: unsold portion of the asset ie. float(0.3) for a sale of 1
                   BTC where another coin of 0.7 was already used
    :param trans: the transaction triggering the report
    :return: a Transaction of type SPLIT with a proper basis
    """
    split_amount_back_in_pool = coin_to_split.quantity_received - amount
    fraction_back_in_pool = split_amount_back_in_pool / coin_to_split.quantity_received
    cost = coin_to_split.quantity_traded * fraction_back_in_pool
    fees = coin_to_split.fees * fraction_back_in_pool
    quantity_received = split_amount_back_in_pool
    to_add = transaction.Transaction(
        transaction.Transaction.Operation.SPLIT,
        symbol_received=coin_to_split.symbol_received,
        quantity_received=quantity_received,
        fees=fees,
        symbol_traded=coin_to_split.symbol_traded,
        quantity_traded=cost,
        date=coin_to_split.date,
    )
    return to_add


def _split_report(coin_to_split, amount, trans, ohlc_provider=None):
    # type:  (transaction.Transaction, Decimal, transaction.Transaction, ohlcprovider.OhlcProvider) -> CostBasisReport
    """
    TODO: make this work for coin/coin trades.
    Given that we are splitting `coin_to_split`, sell`amount` of it and create
    a CBR. The event triggering the CBR is `trans`, typically a sale of some kind.


    The cost basis logic. Note that all fees on buy and sell sides are
    subtracted from the taxable result.

        basis = cost + fees income_subtotal = proceeds - fees taxable_income =
        income_subtotal - basis

    :param coin_to_split: an input. part of this will be the cost basis portion
                          of this report
    :param amount: quantity of the split asset that needs to be sold in this
                   report (not the USD).
    :param trans: the transaction triggering this report, a SELL or SPENDING
    """
    assert amount < coin_to_split.quantity_received
    # TODO: add a test that makes this assertion fail.
    assert not (amount - trans.quantity_traded > 1e-5)  # allowed to be equal
    # coin_to_split is a buy, mining, previous split, some kind of input.
    #               quantity_received: crypto amount.
    #               quantity_traded: USD

    # basis and fee (partial amounts of coin_to_split)
    frac_of_basis_coin = amount / coin_to_split.quantity_received

    # TODO: Lookup missing fiat prices for coin/coin trades.
    purchase_price = frac_of_basis_coin * coin_to_split.quantity_traded
    purchase_fee = frac_of_basis_coin * coin_to_split.fees

    # sale proceeds and fee (again, partial amounts of trans)
    frac_of_sale_tx = amount / trans.quantity_traded
    proceeds = (frac_of_sale_tx * trans.quantity_received).quantize(Decimal(".01"))
    sale_fee = (frac_of_sale_tx * trans.fees).quantize(Decimal(".01"))
    return CostBasisReport(
        trans.user_id,
        purchase_price + purchase_fee,
        amount,
        coin_to_split.date,
        proceeds - sale_fee,
        trans.date,
        trans.symbol_traded,
        triggering_transaction=trans,
    )


def _process_one(trans, pool, ohlc_source=None):
    # type: (transaction.Transaction, coinpool.CoinPool, ohlcprovider.OhlcProvider) -> Sequence
    """
    Cost basis calculator for a single transaction. Return the 'diff'
    required to process this one tx.

    It is assumed that the coin to sell is at the front, at `pool[0]`, so this
    works for both LIFO and FIFO.

    - If transaction is a buy, just return the add-to-pool op.
    - Otherwise, for a sale:
        - Example: buy 0.25 @ $1 and 0.25 at $2. Then sell 0.5 at $3.
                   this is reported to IRS as 2 transactions:
                   One: Sell 0.25 with a basis of $1
                   Two: Sell 0.25 with a basis of $2

    :param trans: a buy or sell with fields filled
    :param pool: a sequence containing transaction.Transaction instances.

    :return (reports, diff, flags): any CostBasisReports, the PoolDiff, any flags raised.
    """
    diff = coinpool.PoolDiff()
    flags = []
    cost_basis_reports = []
    basis_information_absent = False
    amount = Decimal(0)
    pool_index = -1
    if trans.is_simple_input():  # what about: and not trans.is_coin_to_coin():
        diff.add(trans.symbol_received, trans)
        return ([], diff, [])
    # At this point, trans is a sell
    while amount < trans.quantity_traded:
        pool_index += 1
        curr_pool = pool.get(trans.symbol_traded)
        if pool_index >= len(curr_pool):
            # If we get here, we have partial information about the tx.
            # Use a basis of zero for the sale.
            flags.append(("Transaction without basis information", trans))
            basis_information_absent = True
            amount = trans.quantity_traded
        else:
            amount += curr_pool[pool_index].quantity_received
    needs_split = (amount - trans.quantity_traded) > 1e-5

    if needs_split:
        coin_to_split = pool.get(trans.symbol_traded)[pool_index]
        excess = amount - trans.quantity_traded
        portion_of_split_coin_to_sell = coin_to_split.quantity_received - excess
        if trans.is_taxable_output():
            # Outgoing gifts would not trigger this.
            # TODO: Alert the user if the value of a gift exceeds $15,000, in which
            #       case gift taxes may be eligible...
            cost_basis_reports.append(
                _split_report(
                    coin_to_split, portion_of_split_coin_to_sell, trans, ohlc_source
                )
            )
        coin_to_add = _split_coin_to_add(
            coin_to_split, portion_of_split_coin_to_sell, trans
        )
        diff.add(coin_to_add.symbol_received, coin_to_add)
    diff.remove(trans.symbol_traded, pool_index)
    if not needs_split:
        pool_index += 1  # Ensures that we report the oldest transaction as a sale.
    if trans.is_taxable_output():
        # The other option is gift. If it's a gift we don't report any gain or loss.
        # The coins just magically remove themselves from the pool.
        # No entry in 8949 for them.
        cost_basis_reports.extend(
            _build_sale_reports(
                pool, pool_index, trans, basis_information_absent, ohlc_source
            )
        )
    if trans.is_coin_to_coin():
        to_add = _get_coin_to_coin_input(trans)
        diff.add(to_add.symbol_received, to_add)

    return (cost_basis_reports, diff, flags)


def _get_coin_to_coin_input(trans):
    # type: (transaction.Transaction) -> transaction.Transaction
    """
    For a coin/coin trade, we need to add an input back to the pool.
    """
    return transaction.Transaction(
        transaction.Operation.TRADE_INPUT,
        symbol_received=trans.symbol_received,
        quantity_received=trans.quantity_received,
        symbol_traded="USD",
        quantity_traded=0,
        fee_symbol="USD",
        fees=0,
        user_id=trans.user_id,
        source=trans.source,
    )


def _build_sale_reports(pool, pool_index, trans, basis_information_absent, ohlc):
    # type:  (coinpool.CoinPool, int, transaction.Transaction, bool, ohlcprovider.OhlcProvider) -> Sequence[CostBasisReport]
    """
    Use coins from pool to make CostBasisReports.

    :param basis_information_absent: If true, we do not know how much the
        original coin was purchased for yet. This is either because no previous
        buy exists, or because `trans` is a coin/coin trade we need to look up
        the fiat values for.
    :param trans: the tx triggering the reports. It must be a sell of some kind.
    """
    ans = []
    if basis_information_absent:
        report = CostBasisReport(
            trans.user_id,
            decimal.Decimal(0),
            trans.quantity_traded,
            trans.date,
            proceeds=trans.quantity_received,
            date_sold=trans.date,
            asset=trans.symbol_traded,
            triggering_transaction=trans,
        )
        return [report]
    for i in range(pool_index):
        # each of these including pool_index will become a sale to be reported to IRS
        # The cost basis is pool[i].proceeds
        # The sale price is dependent on the parameter `trans'
        #
        # NOTE: the entire basis coin will be sold; but for each iteration
        # through we only use a portion of trans.

        # curr_basis_tx is a BUY, GIFT_RECEIVED or a TRADE_INPUT, or another input.
        curr_basis_tx = pool.get(trans.symbol_traded)[i]
        portion_of_sale = curr_basis_tx.quantity_received / trans.quantity_traded
        # The seller can inflate their cost basis by the buy fees.
        assert curr_basis_tx.symbol_received == trans.symbol_traded
        if not is_fiat(trans.symbol_received):
            if is_fiat(trans.fee_symbol):
                sell_fees = trans.fees
            else:
                sell_fees = ohlc.get(trans.fee_symbol, trans.date).high * trans.fees
            report = CostBasisReport(
                curr_basis_tx.user_id,
                curr_basis_tx.quantity_traded + curr_basis_tx.fees,
                curr_basis_tx.quantity_received,
                date_purchased=curr_basis_tx.date,
                proceeds=portion_of_sale
                * (
                    trans.quantity_received
                    * ohlc.get(trans.symbol_received, trans.date).high
                    - sell_fees
                ),
                date_sold=trans.date,
                asset=trans.symbol_traded,
                secondary_asset=trans.symbol_received,
                triggering_transaction=trans,
            )
        else:
            report = CostBasisReport(
                curr_basis_tx.user_id,
                curr_basis_tx.quantity_traded + curr_basis_tx.fees,
                curr_basis_tx.quantity_received,
                curr_basis_tx.date,
                portion_of_sale * (trans.quantity_received - trans.fees),
                trans.date,
                curr_basis_tx.symbol_received,
                triggering_transaction=trans,
            )
        ans.append(report)
    return ans


def reports_to_csv(reports: Sequence[CostBasisReport]):
    """ Return a file-like object with a row for each report.

    Also includes summary information and headers.
    """
    of = io.StringIO()
    writer = csv.writer(of)
    names = CostBasisReport.field_names()
    writer.writerow(names)
    for r in reports:
        writer.writerow(r.fields())
    writer.writerow(["total", "{}".format(sum([i.gain_or_loss for i in reports]))])
    of.seek(0)
    return of


def _process_all(method, txs, ohlc_source=None):
    # type: (coinpool.PoolMethod, Sequence[Transaction], ohlcprovider.OhlcProvider) -> Sequence
    """
    Create a transaction pool, and iteratively process txs.

    Saves the pool state and any cost basis reports.

    :param method: LIFO or FIFO
    :param txs: a list of transactions (doesn't need to be sorted)
    :return: reports to be sent to tax authorities, and a tx pool.
    """
    assert method in coinpool.PoolMethod
    pool = coinpool.CoinPool(method)
    to_process = sorted(txs, key=lambda trans: trans.date)
    irs_reports = []
    flags = []
    for tx in to_process:
        reports, diff, curr_flags = _process_one(tx, pool, ohlc_source)
        flags.extend(curr_flags)
        pool.apply(diff)
        irs_reports.extend(reports)
    return irs_reports, pool, flags


class BasisProcessor:
    """
    Store state for basis calculations.

    This includes the pool of coins left over at the end of processing the
    given batch.

    See self.pool and self.flags after running process()
    """

    def __init__(self, method, txs, ohlc_class=ohlcprovider.OhlcProvider):
        # type: (coinpool.PoolMethod, Sequence, Callable) -> None
        self.ohlc = ohlc_class()
        self._method = method
        self._txs = txs
        self._reports = []
        self._pool = None
        self._flags = None

    def flags(self):
        if self._flags is None:
            raise RuntimeError("Run basis calculation first")
        return self._flags

    def get_pool(self):
        # type: () -> coinpool.CoinPool
        return self._pool

    def process(self):
        # type: () -> Sequence[CostBasisReport]
        """
        Perform the basis calculation given the txs passed to the constructor.

        Saves the pool of remaining coins and any flags.

        :return: the CostBasisReports generated.
        """
        reports, pool, flags = _process_all(self._method, self._txs, self.ohlc)
        self._reports = reports
        self._pool = pool
        self._flags = flags
        return self._reports



#  Copyright (c) 2019. Seattle Blockchain Solutions. All rights reserved.
#  Licensed under the MIT License. See LICENSE in the project root for license information.
import datetime
import decimal
from collections import namedtuple

OhlcData = namedtuple("OhlcData", ("open", "high", "low", "close"))

jan_1 = datetime.datetime(2017, 1, 1).date()
_ETH_DATA = {
    jan_1: OhlcData(*[decimal.Decimal(i) for i in ["8.5", "8.6", "8.0", "8.1"]])
}
_BTC_DATA = {
    jan_1: OhlcData(*[decimal.Decimal(i) for i in ["1000", "1008.6", "990", "1000"]])
}
_PRICE_DATA = {"ETH": _ETH_DATA, "BTC": _BTC_DATA}


class OhlcProvider:
    """
    Daily OHLC data access stub.
    """

    def __init__(self):
        pass

    def get(self, symbol, dt):
        # type: (str, datetime.datetime) -> OhlcData
        """
        Return the fiat OHLC prices.
        """
        if isinstance(dt, (datetime.datetime)):
            dt = dt.date()
        try:
            val = _PRICE_DATA[symbol][dt]
            return val
        except KeyError:
            raise RuntimeError("No price data found.")



import sqlalchemy
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base(sqlalchemy.schema.ThreadLocalMetaData())

import yabc.server.yabc_api  # isort:skip

bp = yabc.server.yabc_api.bp



# Copyright (c) Seattle Blockchain Solutions. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for license information.
"""
Imports binance exchange data into yabc's exchange-independent format.

yabc is not affiliated with the exchange or company Binance.

TODO: binance is the first supported format with coin/coin exchange defined.
      Once we have historical price data, we can take this live.
"""
import datetime
import decimal
from csv import DictReader
from typing import Sequence

import delorean

from yabc import transaction
from yabc.formats import Format

HEADERS = "Date,Market,Type,Price,Amount,Total,Fee,Fee Coin".split(",")


_BINANCE_TYPE_MAP = {
    "SELL": transaction.Operation.SELL,
    "BUY": transaction.Operation.BUY,
}


def _transaction_from_binance_dict(
    date, market, operation, amount, total, fee, feecoin
):
    # type: (datetime.datetime, str, transaction.Operation, decimal.Decimal, decimal.Decimal, decimal.Decimal, str)-> Sequence

    # The following fields are accurate if the tx is a BUY and not coin/coin.
    symbol_traded = market[:3]
    symbol_received = market[3:]
    quantity_traded = amount
    quantity_received = total
    is_coin_to_coin = not transaction.is_fiat(
        symbol_received
    ) and not transaction.is_fiat(symbol_traded)
    if operation == transaction.Operation.BUY:
        if is_coin_to_coin:
            operation = transaction.Operation.SELL
        # Swap all of these, and the operation
        symbol_received, symbol_traded = (symbol_traded, symbol_received)
        quantity_traded, quantity_received = quantity_received, quantity_traded

    return transaction.Transaction(
        operation=operation,
        source="binance",
        quantity_traded=quantity_traded,
        quantity_received=quantity_received,
        symbol_traded=symbol_traded,
        symbol_received=symbol_received,
        date=date,
        fees=fee,
        fee_symbol=feecoin,
    )


class BinanceParser(Format):
    EXCHANGE_NAME = "Binance"

    def parse(self):
        reader = DictReader(self._file)
        for line in reader:
            date = delorean.parse(line["Date"])
            market = line["Market"]
            operation = _BINANCE_TYPE_MAP[line["Type"]]
            amount = decimal.Decimal(line["Amount"])
            total = decimal.Decimal(line["Total"])
            fee = decimal.Decimal(line["Fee"])
            fee_coin = line["Fee Coin"]
            self._reports.append(
                _transaction_from_binance_dict(
                    date, market, operation, amount, total, fee, fee_coin
                )
            )

    def __init__(self, file=None, filename: str = None):
        self._file = file
        self._reports = []
        self._filename = filename
        self.parse()
        self.cleanup()

    def __next__(self):
        if not self._reports:
            raise StopIteration
        return self._reports.pop(0)

    def __iter__(self):
        return self



"""
Track the sql alchemy session and provide methods for endpoints.
"""
import datetime
import hashlib
import io
import json
import logging
from io import TextIOWrapper

import click
import flask
import sqlalchemy
from flask import make_response
from flask.cli import with_appcontext
from sqlalchemy.orm import sessionmaker

from yabc import Base
from yabc import basis
from yabc import coinpool
from yabc import costbasisreport
from yabc import formats
from yabc import taxdoc
from yabc import transaction
from yabc import user
from yabc.costbasisreport import CostBasisReport
from yabc.formats import coinbase
from yabc.transaction_parser import TransactionParser
from yabc.transaction_parser import TxFile
from yabc.user import User

formats.add_supported_exchanges()

__author__ = "Robert Karl <robertkarljr@gmail.com>"


DB_KEY = "yabc_db"


@click.command("init-db")
@with_appcontext
def init_db_command():
    db = get_db()
    db.create_tables()
    click.echo("Initialized the database.")


@click.command("show-db-config")
@with_appcontext
def show_db_command():
    db = get_db()
    click.echo(db._db_url)


def get_db():
    if DB_KEY not in flask.g:
        flask.g.yabc_db = SqlBackend(flask.current_app.config["DATABASE"])
    return flask.g.yabc_db


def close_db(e=None):
    db = flask.g.pop(DB_KEY, None)
    if db is not None:
        db.session.close()


def init_app(app):
    app.cli.add_command(init_db_command)
    app.cli.add_command(show_db_command)
    app.teardown_appcontext(close_db)


class SqlBackend:
    """
    Handle to the database as well as where DB routines are stored.

    All access to this class should be  done through get_db() above.
    TODO: move get_db() to be a staticmethod here.

    Two backends are supported: sqlite and postgres. Others may work unintentionally.

    NOTE: We must be able to create SqlBackends without a flask instance running.
    """

    def __init__(self, db_url=None):
        if db_url is None:
            db_url = flask.current_app.config["DATABASE"]
        self._db_url = db_url
        # Note: some web servers (aka Flask) will create a new instance of this
        # class for each request.
        self.engine = sqlalchemy.create_engine(
            db_url, poolclass=sqlalchemy.pool.QueuePool
        )
        Session = sessionmaker(bind=self.engine)
        self.session = Session()

    def create_tables(self):
        Base.metadata.create_all(self.engine, checkfirst=True)

    def add_tx(self, userid, tx):
        assert tx
        loaded_tx = coinbase.FromCoinbaseJSON(json.loads(tx))
        loaded_tx.user_id = userid
        self.session.add(loaded_tx)
        self.session.commit()
        val = "Transaction added for user {}. Operation is {}.\n".format(
            userid, loaded_tx.operation
        )
        return val

    def user_create(self, name):
        user_obj = user.User(username=name, password="")
        self.session.add(user_obj)
        self.session.commit()
        return json.dumps(user_obj.id)

    def user_read(self, uid):
        users = self.session.query(user.User).filter_by(id=uid)
        if users.count():
            return json.dumps(users[0])
        return flask.jsonify({"error": "invalid userid"})

    def tx_delete(self, userid, txid):
        """
        Note that we need to clear BasisReports if transactions change.

        :param userid:
        :param txid:
        :return:
        """
        self.session.query(transaction.Transaction).filter_by(
            user_id=userid, id=txid
        ).delete()
        self.session.query(CostBasisReport).filter_by(user_id=userid).delete()
        self.session.commit()
        self.run_basis(userid)

    def tx_delete_by_exchange(self, userid, exchange):
        """
        Deletes CBRs and re-runs basis.
        """
        self.session.query(transaction.Transaction).filter_by(
            user_id=userid, source=exchange
        ).delete()
        self.session.query(CostBasisReport).filter_by(user_id=userid).delete()
        self.session.commit()
        self.run_basis(userid)

    def transactions_clear_all(self, userid):
        """
        Clear all transactions and re-run basis.
        """
        self.session.query(transaction.Transaction).filter_by(user_id=userid).delete()
        self.session.query(CostBasisReport).filter_by(user_id=userid).delete()
        self.session.commit()
        self.run_basis(userid)
        return "success"

    def tx_update(self, userid, txid, values):
        docs = self.session.query(transaction.Transaction).filter_by(
            user_id=userid, id=txid
        )
        obj = docs.all()[0]
        for key in values:
            if key in obj.__dict__:
                setattr(obj, key, values[key])
        self.session.commit()

    def tx_list(self, userid):
        """
        Generate a list of all transactions (buys / sells / gifts) for this user.

        They should be sorted by date.

        :param userid:
        :return: json list, one tx for each item
        """
        txs = (
            self.session.query(transaction.Transaction)
            .filter_by(user_id=userid)
            .order_by(transaction.Transaction.date.asc())
        )
        ans = []
        for tx in txs.all():
            tx_dict = dict(tx.__dict__)
            tx_dict.pop("_sa_instance_state")
            ans.append(tx_dict)
            for numeric_key in (
                "usd_subtotal",
                "fees",
                "quantity",
                "quantity_received",
                "quantity_traded",
            ):
                tx_dict[numeric_key] = str(tx_dict[numeric_key])
            tx_dict["operation"] = tx_dict["operation"].value
        for item in ans:
            item["date"] = str(item["date"])
        return json.dumps(ans)

    def taxdoc_list(self, userid):
        docs = self.session.query(taxdoc.TaxDoc).filter_by(user_id=userid)
        result = []
        for obj in docs.all():
            preview = obj.contents[:10]
            if not isinstance(preview, str):
                # Sqlite seems to return the file contents as bytes, while
                # Postgres returns a string.
                preview = preview.decode()
            result.append(
                {
                    "userid": obj.user_id,
                    "file_name": obj.file_name,
                    "hash": obj.file_hash,
                    "preview": preview + "...",
                }
            )
        return flask.jsonify(result)

    def taxyear_list(self, userid, url_prefix, suffix):
        """
        Basic statistics and download links for each taxyear the given user has data for.

        @param userid
        @param url_prefix (str): for each row, a link for downloading is displayed.
        @param suffix (str): file suffix for the downloaded file.
        """
        current_year = 2019
        user = self.session.query(User).filter_by(id=userid).first()
        sale_dates = self.session.query(
            sqlalchemy.distinct(basis.CostBasisReport.date_sold)
        ).filter_by(user_id=userid)
        years = set([i[0].year for i in sale_dates])
        years.discard(current_year)
        years = sorted(years)
        result = []
        for ty in years:
            year_info = {"year": ty}
            reports = list(self.reports_in_taxyear(userid, ty))
            dollar_keys = ["taxable_income", "shortterm", "longterm"]
            year_info["taxable_income"] = int(sum([i.gain_or_loss for i in reports]))
            year_info["shortterm"] = int(
                sum([i.gain_or_loss for i in reports if not i.long_term])
            )
            year_info["longterm"] = int(
                sum([i.gain_or_loss for i in reports if i.long_term])
            )
            for key in dollar_keys:
                if year_info[key] >= 0:
                    year_info[key] = "${:,}".format(year_info[key])
                else:
                    year_info[key] = "-${:,}".format(-year_info[key])
            year_info["url8949"] = "{}/{}".format(url_prefix, ty)
            year_info["url8949_label"] = "{}-{}-8949.{}".format(
                user.username, ty, suffix
            )
            result.append(year_info)
        return flask.jsonify(result)

    def import_transaction_document(self, userid, submitted_file):
        """
        Add the tx doc for this user.

        - Perform inserts for each of its rows.
        - Recalculate CostBasisReports also.

        @param submitted_file: a filelike object
        exchange and userid should be strings.
        """
        submitted_stuff = submitted_file.read()
        text_mode_file = TextIOWrapper(submitted_file)
        submitted_file.seek(0)
        tx_file = TxFile(text_mode_file, None)
        parser = TransactionParser([tx_file])
        parser.parse()
        if not parser.succeeded():
            val = flask.jsonify({"result": "failure", "flags": parser.flags})
            return make_response(val, 400)
        parsed_txs = parser.txs
        contents_md5_hash = hashlib.md5(submitted_stuff).hexdigest()
        taxdoc_obj = taxdoc.TaxDoc(
            exchange=parser.get_exchange_name(),
            user_id=userid,
            file_hash=contents_md5_hash,
            contents=submitted_stuff,
            file_name=submitted_file.filename,
        )
        self.session.add(taxdoc_obj)
        for tx in parsed_txs:
            tx.user_id = userid
            self.session.add(tx)
        self.session.commit()
        try:
            # It's possible for this to fail if the user uploads documents out of order (sells before buys)
            # TODO: gracefully handle transaction histories where a user reports SELL txs with no basis.
            self.run_basis(
                userid
            )  # TODO: determine if this is a performance bottleneck.
        except Exception as e:
            logging.warn("Failed to run basis ")
            return flask.jsonify({"result": "failure"})
        return flask.jsonify(
            {
                "hash": contents_md5_hash,
                "result": "success",
                "exchange": parser.get_exchange_name(),
                "file_name": taxdoc_obj.file_name,
                "preview": str(submitted_stuff[:20]),
            }
        )

    def run_basis(self, userid):
        """
        Clear any CostBasisReports for this user in the database. Then recalculate them.

        TODO: There is some impedance mismatch between flask and python's csv
            module.  csv requires CSVs to be written as strings, while flask's
            underlying web server requires applications to write binary responses.
            Currently we write to the CSV with strings, and then read the entire
            contents into memory and write to an in-memory binary file-like object
            which is handed off to flask. Fix.

        Returns: just a status
        """
        self.session.query(costbasisreport.CostBasisReport).filter_by(
            user_id=userid
        ).delete()
        all_txs = list(
            self.session.query(transaction.Transaction).filter_by(user_id=userid)
        )
        bp = basis.BasisProcessor(coinpool.PoolMethod.FIFO, all_txs)
        basis_reports = bp.process()

        for i in basis_reports:
            self.session.add(i)
        self.session.commit()
        return "success"

    def reports_in_taxyear(self, userid, taxyear):
        start, end = self.get_tax_year_bounds(userid, taxyear)
        reports = (
            self.session.query(basis.CostBasisReport)
            .filter_by(user_id=userid)
            .filter(basis.CostBasisReport.date_sold >= start)
            .filter(basis.CostBasisReport.date_sold < end)
        )
        return reports

    def download_8949(self, userid, taxyear):
        # type: (int, int) -> io.BytesIO
        reports = self.reports_in_taxyear(userid, taxyear)
        of = basis.reports_to_csv(reports)
        mem = io.BytesIO()
        mem.write(of.getvalue().encode("utf-8"))
        mem.seek(0)
        return mem

    def get_tax_year_bounds(self, userid, taxyear):
        first_invalid_date = datetime.datetime(taxyear + 1, 1, 1)
        start = datetime.datetime(taxyear, 1, 1)
        return start, first_invalid_date



import csv
import decimal
import enum

import delorean

from yabc import transaction
from yabc.formats import FORMAT_CLASSES
from yabc.formats import Format
from yabc.transaction import Operation


"""
TXID, Created, Received, Sent, TXtype, TXdesc, TXNotes
,2015-04-05T02:03:47+00:00,1.50000000,,Deposit,
"""


class LocalBitcoinTradeTypes(enum.Enum):
    # Sales (for tax purposes)
    LOCAL_SELL = "LOCAL_SELL"
    ONLINE_BUY = "ONLINE_BUY"
    # Purchasing bitcoin (for tax purposes)
    ONLINE_SELL = "ONLINE_SELL"
    LOCAL_BUY = "LOCAL_BUY"


def _is_sell(trade_type: LocalBitcoinTradeTypes):
    return trade_type in {
        LocalBitcoinTradeTypes.LOCAL_SELL,
        LocalBitcoinTradeTypes.ONLINE_BUY,
    }


class LocalBitcoinsParser(Format):
    EXCHANGE_NAME = "localbitcoins"

    def attempt_read_transaction(self, line):
        try:
            kind = LocalBitcoinTradeTypes[line["trade_type"]]
            date = delorean.parse(
                line["transaction_released_at"], dayfirst=False
            ).datetime
            btc_amount = line["btc_traded"]
            fiat = decimal.Decimal(line["fiat_amount"])
            fiat_fee = decimal.Decimal(line["fiat_fee"])
            if _is_sell(kind):
                tx_type = Operation.SELL
            else:
                tx_type = Operation.BUY
            return transaction.Transaction(
                operation=tx_type,
                asset_name="BTC",
                date=date,
                fees=fiat_fee,
                quantity=btc_amount,
                source=LocalBitcoinsParser.EXCHANGE_NAME,
                usd_subtotal=fiat,
            )
        except RuntimeError:
            raise RuntimeError("Could not parse localbitcoins data.")
        except KeyError as e:
            raise RuntimeError("Unknown key in localbitcoins file: {}".format(e))

    def __init__(self, csv_content=None, filename=None):
        if csv_content:
            assert not filename
            csv_content.seek(0)
            self._file = csv_content
            self._reader = csv.DictReader(csv_content)
        else:
            assert not csv_content
            self._file = open(filename, "r")
            self._reader = csv.DictReader(self._file)
        self.txs = []
        for line in self._reader:
            tx = self.attempt_read_transaction(line)
            self.txs.append(tx)

    def __iter__(self):
        return self

    def __next__(self):
        if not self.txs:
            raise StopIteration
        return self.txs.pop(0)


FORMAT_CLASSES.append(LocalBitcoinsParser)



import datetime
import decimal
from collections import defaultdict
from decimal import Decimal

import sqlalchemy
from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import orm

import yabc
from yabc import transaction
from yabc.transaction import PreciseDecimalString


class CostBasisReport(yabc.Base):
    """
    Represents a row in form 8949.
    TODO: Check that long term results match expected in leap years.

    >>> r = CostBasisReport.make_random_report()
    >>> r.date_sold = datetime.datetime(2019,1,1)
    >>> r.date_purchased = datetime.datetime(2018,1,1)
    >>> r._is_long_term()
    True
    >>> r.date_purchased = datetime.datetime(2018,1,2)
    >>> r._is_long_term()
    False
    """

    _fields = [
        "id",
        "asset_name",
        "basis",
        "date_purchased",
        "date_sold",
        "proceeds",
        "quantity",
        "long_term",
        "adjustment",
        "user_id",
    ]

    __tablename__ = "basis_report"
    id = Column(Integer, primary_key=True)
    asset_name = Column(sqlalchemy.String)
    basis = Column(PreciseDecimalString)
    date_purchased = Column(DateTime)
    date_sold = Column(DateTime)
    proceeds = Column(PreciseDecimalString)
    quantity = Column(PreciseDecimalString)
    adjustment = Column(PreciseDecimalString)
    long_term = Column(Boolean)
    user_id = Column(sqlalchemy.Integer, ForeignKey("user.id"))
    secondary_asset = Column(sqlalchemy.String)  # Needed for coin/coin trades.

    @staticmethod
    def make_random_report():
        return CostBasisReport(
            1,
            Decimal(1),
            Decimal(1),
            datetime.datetime.now(),
            Decimal(2),
            datetime.datetime.now(),
            "BTC",
        )

    def __init__(
        self,
        userid,
        basis,
        quantity,
        date_purchased,
        proceeds,
        date_sold,
        asset,
        adjustment=Decimal(0),
        triggering_transaction=None,
        secondary_asset=None,
    ):
        # type: (int, decimal.Decimal, decimal.Decimal, datetime.datetime, decimal.Decimal, datetime.datetime, str, decimal.Decimal, transaction.Transaction) -> None
        """
        Note that when pulling items from a SQL alchemy ORM query, this constructor isn't called.
        """
        assert isinstance(date_sold, datetime.datetime)
        assert isinstance(date_purchased, datetime.datetime)
        assert isinstance(asset, str)
        self._round_to_dollar = True
        self.user_id = userid
        self.basis = self._round_as_needed(basis)
        self.quantity = quantity
        self.date_purchased = date_purchased
        self.proceeds = self._round_as_needed(proceeds)
        self.date_sold = date_sold
        self.asset_name = asset
        self.adjustment = adjustment
        self.long_term = self._is_long_term()
        self.triggering_transaction = triggering_transaction
        self.secondary_asset = secondary_asset

    def _is_long_term(self):
        return (self.date_sold - self.date_purchased) > datetime.timedelta(364)

    def __getattr__(self, item):
        if item == "gain_or_loss":
            return self.get_gain_or_loss(True)
        raise AttributeError()

    def get_gain_or_loss(self, round_dollars=True):
        """
        This field is calculated based on proceeds and basis.

        It could return different values based on different printings of a user's tax info.

        :param round_dollars: We only support True so far.
        :return:
        """
        return self._round_as_needed(self.proceeds - self.basis)

    def _round_as_needed(self, amount: decimal.Decimal):
        if self._round_to_dollar:
            return amount.quantize(1)
        return amount.quantize(Decimal(".01"))

    def description(self):
        """
        A description suitable for form 8949.

        For example, .06 BTC to USD
        """
        return "{:.6f} {} {}".format(
            self.quantity,
            self.asset_name,
            "to USD"
            if not self.secondary_asset
            else "to {}".format(self.secondary_asset),
        )

    def __repr__(self):
        return "<Sold {} {} on {date} for ${}. Exchange: {exchange}. Profit:${profit}.{longterm}\n\t{trigger}>".format(
            self.quantity,
            self.asset_name,
            self.proceeds,
            profit=self.gain_or_loss,
            date=self.date_sold,
            longterm=" Long term." if self.long_term else "",
            exchange=self.triggering_transaction.source
            if self.triggering_transaction
            else "Unknown",
            trigger=self.triggering_transaction,
        )

    def fields(self):
        ans = []
        for i in CostBasisReport._fields:
            ans.append(getattr(self, i))
        return ans

    @staticmethod
    def field_names():
        return CostBasisReport._fields

    @orm.reconstructor
    def init_on_load(self):
        self._round_to_dollar = True


class ReportBatch:
    """
    Contain a list of reports and any context needed for calculating totals.

    8949 needs the total revenue, for example.

    Most notable is whether or not to round to the nearest dollar.

    >>> r1 = CostBasisReport.make_random_report()
    >>> r1.gain_or_loss = decimal.Decimal('1.5')
    >>> r2 = CostBasisReport.make_random_report()
    >>> r2.gain_or_loss = decimal.Decimal('1.5')
    >>> batch = ReportBatch([r1, r2], True)
    >>> batch.totals()['gain_or_loss'] == decimal.Decimal('4')
    True
    >>> batch_no_rounding = ReportBatch([r1, r2], False)
    >>> batch_no_rounding.totals()['gain_or_loss'] == decimal.Decimal('3')
    True
    """

    KEYS = ["proceeds", "basis", "adjustment", "gain_or_loss"]

    def __init__(self, reports, round_dollars=True):
        self.reports = reports
        self.round_dollars = round_dollars
        self._totals = None

    def totals(self):
        """
        :return:  dict with keys ["proceeds", "basis", "adjustment", "gain_or_loss"]
        """
        if not self._totals:
            self._totals = defaultdict(decimal.Decimal)
            for r in self.reports:
                for key in self.KEYS:
                    self._totals[key] += getattr(r, key)
        return self._totals

    def human_readable_report(self):
        """
        Given a list of CostBasisReports to be submitted to tax authorities, generate a human
        readable report.

        :return str:
        """
        ans = ""
        ans += "{} transactions to be reported\n\n".format(len(self.reports))
        for i in self.reports:
            ans += "{}\n".format(str(i))
        ans += "\ntotal gain or loss for above transactions: {}".format(
            self.totals()["gain_or_loss"]
        )
        ans += "\n"
        ans += "\ntotal basis for above transactions: {}".format(self.totals()["basis"])
        ans += "\ntotal proceeds for above transactions: {}".format(
            self.totals()["proceeds"]
        )
        return ans



# Copyright (c) Seattle Blockchain Solutions. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for license information.
import logging


class Format:
    """
    Base class for formats from various exchanges.
    """

    EXCHANGE_NAME = "Unknown exchange"

    @classmethod
    def exchange_name(cls):
        """
        Human-friendly name of this exchange.
        :return: str
        """
        return cls.EXCHANGE_NAME

    def has_next(self):
        if not self._reports:
            return False
        return True

    def cleanup(self):
        logging.info("cleaning up file {}".format(self._file))
        if self._file:
            try:
                # self._file can be a sequence, which shouldn't have close().
                self._file.close()
            except AttributeError:
                pass
            del self._file


FORMAT_CLASSES = []


def add_supported_exchanges():
    """
    This is in a function mostly so our automated reformatting scripts don't place these imports above all statements

    #noqa should stop our unused import zapper from zapping.
    """
    # TODO: these are circular imports; these files then attemp to import THIS file.
    from . import adhoc  # noqa
    from . import coinbase  # noqa
    from . import gemini  # noqa
    from . import localbitcoins  # noqa



import functools
import os

import flask
from flask import Blueprint

from yabc import user
from yabc.server import sql_backend
from yabc.user import User

yabc_api = Blueprint("yabc_api", __name__)
bp = yabc_api
USER_ID_KEY = "user_id"


def is_authorized(userid):
    """ @param userid: needs to match the logged in user. """
    FLASK_ENV_KEY = "FLASK_ENV"
    env = os.environ.get(FLASK_ENV_KEY)
    if env == "development":
        # Anything goes in development.
        return True
    if not flask.g.user:
        return False
    assert isinstance(flask.g.user, user.User)
    if flask.g.user.id == userid:
        return True
    return False


def check_authorized(view):
    """
    Check the userid attached to the request against the logged-in user.
    """

    @functools.wraps(view)
    def wrapped_view(**kwargs):
        userid = flask.request.args.get(USER_ID_KEY)
        if not userid:
            userid = flask.session.get(USER_ID_KEY)
        if not userid:
            raise RuntimeError("No userid, not authorized")
        if not is_authorized(userid):
            return flask.make_response(("", 500))
        return view(**kwargs)

    return wrapped_view


def get_userid():
    userid = flask.request.args.get(USER_ID_KEY)
    if not userid:
        userid = flask.session.get(USER_ID_KEY)
    if not userid:
        raise RuntimeError("User is not logged in. get_userid() failed.")
    return userid


@yabc_api.route("/yabc/v1/run_basis", methods=["POST"])
@check_authorized
def run_basis():
    """
    Perform the cost basis calculations and write them all to the database.
    """
    userid = get_userid()
    backend = sql_backend.get_db()
    result = backend.run_basis(userid)
    return result


@yabc_api.route("/yabc/v1/download_8949/<taxyear>", methods=["GET"])
@check_authorized
def download_8949(taxyear):
    """
    Get the relevant tax document for a given year.
    """
    userid = get_userid()
    backend = sql_backend.get_db()
    user = backend.session.query(User).filter_by(id=userid).first()
    of = backend.download_8949(userid, int(taxyear))
    result = flask.send_file(
        of,
        mimetype="text/csv",
        cache_timeout=1,
        attachment_filename="{}-{}-8949.csv".format(user.username, taxyear),
        as_attachment=True,
    )
    return result


@yabc_api.route("/yabc/v1/taxyears", methods=["GET"])
@check_authorized
def taxyears():
    """
    No backend data structure corresponds to this endpoint;

    It's client-friendly condensed information from the CostBasisReport table.

    Each year that has tax information is included.
    """
    userid = get_userid()
    backend = sql_backend.get_db()
    return backend.taxyear_list(userid)


@yabc_api.route("/yabc/v1/taxdocs", methods=["POST", "GET"])
@check_authorized
def taxdocs():
    userid = get_userid()
    backend = sql_backend.get_db()
    if flask.request.method == "GET":
        return backend.taxdoc_list(userid)
    submitted_file = flask.request.files["taxdoc"]
    return backend.import_transaction_document(userid, submitted_file)


@yabc_api.route("/yabc/v1/transactions/by-exchange/<exchange_name>", methods=["DELETE"])
@check_authorized
def delete_by_exchange(exchange_name):
    userid = get_userid()
    try:
        sql_backend.get_db().tx_delete_by_exchange(userid, exchange_name)
        sql_backend.close_db()
    except RuntimeError:
        return flask.make_response(("Failed to delete.", 500))
    return flask.jsonify(
        {"result": "Deleted transactions for exchange {}".format(exchange_name)}
    )


@yabc_api.route("/yabc/v1/transactions/<txid>", methods=["DELETE"])
@check_authorized
def transaction_update(txid):
    userid = get_userid()
    backend = sql_backend.get_db()
    if flask.request.method == "DELETE":
        backend.tx_delete(userid, txid)
    elif flask.request.method == "PUT":
        backend.tx_update(userid, txid, flask.request.values)
    else:
        raise ValueError(
            "method {} not support for transaction".format(flask.request.method)
        )
    sql_backend.close_db()
    return flask.jsonify({"result": "Deleted transaction with id {}".format(txid)})


@yabc_api.route("/yabc/v1/transactions", methods=["GET", "POST", "DELETE"])
@check_authorized
def transactions():
    userid = get_userid()
    backend = sql_backend.get_db()
    if flask.request.method == "GET":
        return backend.tx_list(userid)
    elif flask.request.method == "DELETE":
        return backend.transactions_clear_all(userid)
    tx = flask.request.values["tx"]
    assert tx
    return backend.add_tx(userid, tx)


@yabc_api.route("/yabc/v1/users/<userid>", methods=["GET"])
@check_authorized
def user_read(userid):
    backend = sql_backend.get_db()
    return backend.user_read(userid)


@yabc_api.route("/yabc/v1/users", methods=["POST"])
def user_create():
    name = flask.request.args.get("username")
    backend = sql_backend.get_db()
    return backend.user_create(name)



from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import String

import yabc


class User(yabc.Base):
    __tablename__ = "user"
    id = Column(Integer, primary_key=True)
    username = Column(String, nullable=False, unique=True)
    password = Column(String)

    def __init__(self, username, password):
        self.username = username
        self.password = password



"""
Entry point to command line yabc usage.

Example:

```
pip install yabc
python -m yabc statement_from_coinbase.csv my_gemini_statement.csv adhoc.csv
```

File types are automatically detected.
"""
import argparse
import sys

import yabc.transaction_parser
from yabc import basis
from yabc import coinpool
from yabc.costbasisreport import ReportBatch


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("filenames", nargs="+", metavar="filename")
    args = parser.parse_args()
    tx_files = [
        yabc.transaction_parser.TxFile(open(fname), None) for fname in args.filenames
    ]
    parser = yabc.transaction_parser.TransactionParser(tx_files)
    parser.parse()
    if parser.flags:
        for flag in parser.flags:
            print(flag, file=sys.stderr)
        print("Quitting yabc.", file=sys.stderr)
        sys.exit(1)
    processor = basis.BasisProcessor(coinpool.PoolMethod.FIFO, parser.txs)
    reports = processor.process()
    batch = ReportBatch(reports)
    print(batch.human_readable_report())
    print("Remaining coins after sales:")
    for symbol in processor.get_pool().known_symbols():
        for coin in processor.get_pool().get(symbol):
            print(coin)


if __name__ == "__main__":
    main()



"""
Adhoc transactions - yabc's proprietary format for specifying trades, mining,
espending, and gifts.
"""
import csv
import datetime
import decimal

import delorean

from yabc import transaction
from yabc.formats import FORMAT_CLASSES
from yabc.formats import Format

_TRANSACTION_TYPE_HEADER = "Type"
_RECEIVED_CURRENCY = "ReceivedCurrency"
_TIMESTAMP_HEADER = "Timestamp"
_RECEIVED_AMOUNT = "ReceivedAmount"
_TRADED_CURRENCY = "TradedCurrency"
_TRADED_AMOUNT = "TradedAmount"
_FEE = "Fee"
_FEE_COIN = "FeeCurrency"

_FIELD_NAMES = [
    _RECEIVED_CURRENCY,
    _RECEIVED_AMOUNT,
    _TIMESTAMP_HEADER,
    _TRANSACTION_TYPE_HEADER,
    _TRADED_CURRENCY,
    _TRADED_AMOUNT,
    _FEE,
    _FEE_COIN,
]

_SUPPORTED = [
    transaction.Transaction.Operation.GIFT_RECEIVED,
    transaction.Transaction.Operation.GIFT_SENT,
    transaction.Transaction.Operation.MINING,
    transaction.Transaction.Operation.SPENDING,
    transaction.Transaction.Operation.SELL,
    transaction.Transaction.Operation.BUY,
]


class AdhocParser(Format):
    """
    Defines an input format for ad-hoc transactions like mining, spending, and gifts.

    This class translates CSV rows into transaction objects.
    """

    EXCHANGE_NAME = "adhoc"

    def validate_headers(self, curr):
        for header_name in _FIELD_NAMES:
            if header_name not in curr:
                raise RuntimeError(
                    "Incorrectly formatted adhoc file {}, missing header key {}".format(
                        self._file, header_name
                    )
                )

    def __init__(self, csv_file):
        """
        :param csv_file: can be a list of rows, each containing a string, or an
        open file-like object.
        """
        self._file = None
        assert not isinstance(csv_file, str)
        if isinstance(csv_file, (list, tuple)):
            self.validate_headers(csv_file[0])
        else:
            csv_file.seek(0)
            contents = []
            while True:
                line = csv_file.readline()
                if line:
                    contents.append(line)
                else:
                    break
            rawcsv = list(csv.reader(contents))
            if not rawcsv:
                raise RuntimeError("not enough rows in adhoc file {}".format(csv_file))
            fieldnames = rawcsv[0]
            self.validate_headers(fieldnames)
            csv_file.seek(0)
        self._file = csv_file
        self.reader = csv.DictReader(self._file)

    def __next__(self):
        curr = next(self.reader)
        if not curr:
            raise StopIteration
        op = None
        for t in _SUPPORTED:
            if curr[_TRANSACTION_TYPE_HEADER].upper() == t.value.upper():
                op = t
        if not op:
            return transaction.Transaction(operation=transaction.Operation.NOOP)
        trans_date = delorean.parse(curr[_TIMESTAMP_HEADER], dayfirst=False).datetime

        if op == transaction.Operation.MINING:
            return _handle_mining(trans_date, curr)
        elif op == transaction.Operation.GIFT_RECEIVED:
            return _handle_gift_received(trans_date, curr)
        elif op == transaction.Operation.SPENDING:
            return _handle_spending(trans_date, curr)
        elif op == transaction.Operation.GIFT_SENT:
            return _handle_gift_sent(trans_date, curr)
        elif op == transaction.Operation.SELL:
            return _make_adhoc_sell(trans_date, curr)
        elif op == transaction.Operation.BUY:
            return _handle_buy(trans_date, curr)
        else:
            # Should be unreachable as we checked if we did not match above.
            assert False

    def __iter__(self):
        return self


def _read_usd_value(usd_str):
    # type: (str) -> decimal.Decimal
    """
    Accept '$10,000.00' '$10,000' or '10000' or ''
    """
    if not usd_str:
        return None
    usd_str = usd_str.lstrip("$")
    usd_str = usd_str.replace(",", "")
    if not usd_str:
        return decimal.Decimal(0)
    return decimal.Decimal(usd_str)


def _handle_spending(date, curr):
    # type: (datetime.datetime, dict) -> transaction.Transaction
    cost_basis = 0
    fees = 0
    fee_coin = "USD"
    if curr[_RECEIVED_AMOUNT]:
        cost_basis = _read_usd_value(curr[_RECEIVED_AMOUNT])
    if curr[_FEE]:
        fees = curr[_FEE]
        fee_coin = curr[_FEE_COIN]

    return transaction.Transaction(
        transaction.Operation.SPENDING,
        date=date,
        symbol_traded=curr[_TRADED_CURRENCY],
        quantity_traded=decimal.Decimal(curr[_TRADED_AMOUNT]),
        quantity_received=cost_basis,
        symbol_received="USD",
        fees=fees,
        fee_symbol=fee_coin,
        source=AdhocParser.EXCHANGE_NAME,
    )


def _handle_buy(date, curr):
    # type: (datetime.datetime, dict) -> transaction.Transaction
    fees = 0
    feecoin = "USD"
    if curr[_FEE]:
        fees = curr[_FEE]
        feecoin = curr[_FEE_COIN]
    return transaction.Transaction(
        transaction.Operation.BUY,
        date=date,
        symbol_traded="USD",
        quantity_traded=curr[_TRADED_AMOUNT],
        symbol_received=curr[_RECEIVED_CURRENCY],
        quantity_received=curr[_RECEIVED_AMOUNT],
        fees=fees,
        fee_symbol=feecoin,
        source=AdhocParser.EXCHANGE_NAME,
    )


def _handle_gift_received(date, curr):
    cost_basis = 0
    basis_currency = "USD"
    if curr[_TRADED_AMOUNT]:
        cost_basis = _read_usd_value(curr[_TRADED_AMOUNT])

    return transaction.Transaction(
        transaction.Operation.GIFT_RECEIVED,
        date=date,
        # Cost basis, optional
        symbol_traded=basis_currency,
        quantity_traded=cost_basis,
        # What was actually received
        quantity_received=curr[_RECEIVED_AMOUNT],
        symbol_received=curr[_RECEIVED_CURRENCY],
        source=AdhocParser.EXCHANGE_NAME,
    )


def _handle_gift_sent(date, curr):
    return transaction.Transaction(
        transaction.Operation.GIFT_SENT,
        date=date,
        symbol_traded=curr[_TRADED_CURRENCY],
        quantity_traded=curr[_TRADED_AMOUNT],
        symbol_received="USD",
        quantity_received=0,
        source=AdhocParser.EXCHANGE_NAME,
    )


def _handle_mining(date, curr):
    # type: (datetime.datetime, dict) -> transaction.Transaction
    # TODO: for mining we need to look up the value on the date mined.
    #       Optionally, the value can be specified
    mined_value = 0
    if curr[_TRADED_AMOUNT]:
        mined_value = _read_usd_value(curr[_TRADED_AMOUNT])

    trans = transaction.Transaction(
        operation=transaction.Operation.MINING,
        quantity_received=curr[_RECEIVED_AMOUNT],
        symbol_received=curr[_RECEIVED_CURRENCY],
        fees=0,
        symbol_traded="USD",
        quantity_traded=mined_value,
        date=date,
        source=AdhocParser.EXCHANGE_NAME,
    )
    return trans


def _make_adhoc_sell(date, curr):
    if curr[_RECEIVED_CURRENCY]:
        fee = 0
        fee_coin = "USD"
        rcvd_coin = curr[_RECEIVED_CURRENCY]
        rcvd_amount = curr[_RECEIVED_AMOUNT]
        if not rcvd_amount or not rcvd_coin:
            raise RuntimeError(
                "Invalid row, need currency and amount if either is specified {}".format(
                    curr
                )
            )
        if curr[_FEE]:
            fee = decimal.Decimal(curr[_FEE])
            fee_coin = curr[_FEE_COIN]
        return transaction.Transaction(
            transaction.Operation.SELL,
            symbol_received=rcvd_coin,
            quantity_received=decimal.Decimal(rcvd_amount),
            symbol_traded=curr[_TRADED_CURRENCY],
            quantity_traded=curr[_TRADED_AMOUNT],
            date=date,
            fees=fee,
            fee_symbol=fee_coin,
            source=AdhocParser.EXCHANGE_NAME,
        )


FORMAT_CLASSES.append(AdhocParser)



from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import String

import yabc


class TaxDoc(yabc.Base):

    """
    TODO: If we need to store these, do so in an encrypted object storage, not as rows in a databse.
    """

    __tablename__ = "taxdoc"
    id = Column(Integer, primary_key=True)
    user_id = Column(Integer, ForeignKey("user.id"))
    file_name = Column(String)
    file_hash = Column(String)
    exchange = Column(String)
    contents = Column(String)



#  Copyright (c) 2019. Seattle Blockchain Solutions. All rights reserved.
#  Licensed under the MIT License. See LICENSE in the project root for license information.
import enum
from collections import defaultdict
from typing import Sequence

from yabc import transaction


@enum.unique
class PoolMethod(enum.Enum):
    LIFO = 1
    FIFO = 2


class PoolDiff:
    """
    Operations to be applied to a CoinPool.

    Supports add and remove.
    """

    def __init__(self):
        self.to_add = []
        self.to_remove = []

    def add(self, symbol, coin):
        # type: (str, transaction.Transaction) -> None
        if not coin.symbol_received == symbol:
            raise RuntimeError("Cannot add to pool {}".format(symbol))
        self.to_add.append(coin)

    def remove(self, symbol, index):
        # type: (str, int) -> None
        """
        The coins up to and including this index will be REMOVED from the pool.
        """
        self.to_remove.append((symbol, index))


def _handle_add_lifo(pool, to_add: transaction.Transaction):
    """
    Simply put any new transaction, including splits, at the beginning.
    """
    pool.insert(0, to_add)


def _handle_add_fifo(pool, to_add: transaction.Transaction):
    """
    FIFO is defined by putting the BUY transactions at the end.
    For split coins, they need to be sold first.
    """
    if to_add.operation == transaction.Operation.SPLIT:
        pool.insert(0, to_add)
    else:
        assert to_add.operation in [
            transaction.Operation.BUY,
            transaction.Operation.MINING,
            transaction.Operation.GIFT_RECEIVED,
            transaction.Operation.TRADE_INPUT,
        ]
        pool.append(to_add)


class CoinPool:
    """
    Consider trading BTC, getting ETH, and paying the fee in ETH.

    - need to add ETH to the pool, less fees
    - need to remove BTC from the pool
        - potentially need to split a BTC coin
        - if so, need to add BTC to the pool

    Internal representation note: by convention, the coins at the START of the
    list will be sold first. This is for LIFO and FIFO.
    """

    def __init__(self, method):
        # type: (PoolMethod) -> None
        assert method in PoolMethod
        self._coins = defaultdict(list)
        self.method = method

    def known_symbols(self):
        return self._coins.keys()

    def get(self, coin_name):
        # type: (str) -> Sequence[transaction.Transaction]
        """
        Read-only access to the pool of transactions for a given symbol.

        Note that it's read-only by convention, and clients who modify the returned value are in for trouble.
        TODO: Consider returning a copy here (?)

        :return: A sequence of Transaction
        """
        return self._coins[coin_name]

    def apply(self, diff: PoolDiff):
        """
        Pop from the front of a list of txs, or add in the correct spot based on method.

        This mutates the object and isn't easy to undo.
        """
        for item in diff.to_add:
            coin_list = self._coins[item.asset_name]
            if self.method == PoolMethod.LIFO:
                _handle_add_lifo(coin_list, item)
            elif self.method == PoolMethod.FIFO:
                _handle_add_fifo(coin_list, item)
        for symbol, index in diff.to_remove:
            self._coins[symbol] = self._coins[symbol][index + 1 :]



import collections
import logging
from typing import Sequence

import yabc
import yabc.formats

TxFile = collections.namedtuple("InputFile", ("file", "parser_hint"))


class TransactionParser:
    """
    Turn collections of file-like objects into a transaction list.
    """

    def parse(self):
        for f, hinted_parser in self.files:
            self.txs.extend(self._get_txs(f, hinted_parser))

    def __init__(self, files: Sequence[TxFile]):
        """

        :param files: a sequence of TxFile objects.
        """
        self.txs = []
        self.flags = []
        self.files = files
        self._exchange = None
        self._success = False

    def get_exchange_name(self):
        return self._exchange.exchange_name()

    def succeeded(self):
        return self._success

    def _get_txs(self, f, hinted_parser):
        if hinted_parser is not None:
            return list(hinted_parser(f))
        # Try formats until something works.
        for constructor in yabc.formats.FORMAT_CLASSES:
            try:
                generator = constructor(f)
                values = list(generator)
                self._exchange = constructor
                self._success = True
                return values
            except RuntimeError as e:
                logging.info(e)
                continue
        self._success = False
        self.flags.append("Couldn't find any transactions in file.")
        return []



#!/usr/bin/env python3


import setuptools

"""
Python versions and dependencies notes:

- dateutil for 3.4 doesn't have parser accessible on the package
- previously there was a circular import issue in our formats package. This
  failed on 3.4 (but worked on later versions)
- we need the typing dep since we need the backport for 3.4
"""

setuptools.setup(
    version="0.1.13",
    name="yabc",
    python_requires=">=3.4,<=3.8",
    author="Seattle Blockchain Solutions",
    maintainer="Robert Karl",
    maintainer_email="robertkarljr@gmail.com",
    install_requires=[
        "flask==1.0.4",  # flask 1.1. drops python3.4 support
        "sqlalchemy==1.3.3",
        "delorean==1.0.0",
        "python-dateutil",  # TODO: remove, we can just use delorean
        # This package is present on 3.5+, backport required for 3.4
        "typing",
    ],
    url="https://github.com/robertkarl/yabc",
    description="A tax estimator for cryptocurrencies.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3 :: Only",
    ],
)



"""
TODO: Accept Gemini .xlsx files. Do not require the user to convert to CSV.
"""
import csv
import decimal

import delorean

from yabc import formats
from yabc import transaction

_CURRENCIES = ["BCH", "BTC", "ZEC", "ETH", "LTC"]
_SOURCE_NAME = "gemini"
_TIME_HEADER = "Time (UTC)"
_TYPE_HEADER = "Type"


def _gem_int_from_dollar_string(s):
    s = s.strip(" $()")
    s = s.replace(",", "")
    return decimal.Decimal(s)


def _quantity(tx_row):
    """
    :return: a decimal.Decimal with the quantity of eth, btc, etc traded.
    """
    # Currency field looks like 'BTCUSD'
    currency = tx_row["Symbol"].rstrip("USD")
    if currency not in _CURRENCIES:
        raise RuntimeError("Currency {} not supported!".format(currency))

    column_name = "{} Amount {}".format(currency, currency)
    return (decimal.Decimal(tx_row[column_name].strip("(").split(" ")[0]), currency)


def _tx_from_gemini_row(tx_row):
    """
    Create a Transaction based on a row. May return None.

    :param tx_row: a dictionary with keys from a gemini transaction history spreadsheet.
    :return:  None if not a transaction needed for taxes. Otherwise a Transaction object.
    """
    if _TYPE_HEADER not in tx_row:
        raise RuntimeError("Not a valid gemini file.")
    if tx_row[_TYPE_HEADER] not in ("Buy", "Sell"):
        return None
    date = delorean.parse(
        "{} {}".format(tx_row["Date"], tx_row[_TIME_HEADER]), dayfirst=False
    ).datetime
    usd_subtotal = _gem_int_from_dollar_string(tx_row["USD Amount USD"])
    fees = _gem_int_from_dollar_string(tx_row["Fee (USD) USD"])
    quantity, currency = _quantity(tx_row)
    quantity = quantity
    tp = _gemini_type_to_operation(tx_row["Type"])
    if tp == transaction.Operation.BUY:
        tx = transaction.Transaction(
            operation=tp,
            quantity_traded=usd_subtotal,
            symbol_traded="USD",
            quantity_received=quantity,
            symbol_received=currency,
            date=date,
            fees=fees,
            source=_SOURCE_NAME,
        )
    elif tp == transaction.Operation.SELL:
        tx = transaction.Transaction(
            operation=tp,
            quantity_traded=quantity,
            symbol_traded=currency,
            quantity_received=usd_subtotal,
            symbol_received="USD",
            date=date,
            fees=fees,
            source=_SOURCE_NAME,
        )
    else:
        raise RuntimeError("Unknown transaction type from gemini file.")
    return tx


def _read_txs_from_file(f):
    """
    Validate headers and read buy/sell transactions from the open file-like object 'f'.

    Note: we use the seek method on f.
    """
    f.seek(0)
    rawcsv = [i for i in csv.reader(f)]
    if not rawcsv:
        raise RuntimeError("not enough rows in gemini file {}".format(f))
    fieldnames = rawcsv[0]
    _valid_gemini_headers(fieldnames)
    f.seek(0)
    transactions = [i for i in csv.DictReader(f, fieldnames)][1:]
    ans = []
    for row in transactions:
        item = _tx_from_gemini_row(row)
        if item is not None:
            ans.append(item)
    return ans


def _valid_gemini_headers(fieldnames):
    """
    Make sure we have the required headers to be sure this is a gemini file.
    """
    required_fields = "Type,Date,Time (UTC),Symbol,BTC Amount BTC,USD Amount USD,Fee (USD) USD".split(
        ","
    )
    for field in required_fields:
        if field not in fieldnames:
            raise RuntimeError(
                "Not a valid gemini file. Requires header '{}'".format(field)
            )


class GeminiParser(formats.Format):
    def __init__(self, fname_or_file):
        self.txs = []
        self.flags = []
        self._file = fname_or_file
        if isinstance(fname_or_file, str):
            with open(fname_or_file) as f:
                self.txs = _read_txs_from_file(f)
        else:
            # it must be an open file
            self.txs = _read_txs_from_file(fname_or_file)

    EXCHANGE_NAME = "Gemini"

    def __iter__(self):
        return self

    def __next__(self):
        if not self.txs:
            raise StopIteration
        return self.txs.pop(0)


def _gemini_type_to_operation(gemini_type: str):
    if gemini_type == "Buy":
        return transaction.Transaction.Operation.BUY
    elif gemini_type == "Sell":
        return transaction.Transaction.Operation.SELL
    raise RuntimeError(
        "Invalid Gemini transaction type {} encountered.".format(gemini_type)
    )


formats.FORMAT_CLASSES.append(GeminiParser)



"""
Definition of a Transaction, the in-memory version of an asset buy/sell
"""
import datetime
import enum
from decimal import Decimal

import sqlalchemy.orm
from sqlalchemy.types import TypeDecorator

import yabc

__author__ = "Robert Karl <robertkarljr@gmail.com>"


class PreciseDecimalString(TypeDecorator):
    impl = sqlalchemy.String

    def process_bind_param(self, value, dialect):
        """ Needs to return an object of the underlying impl 
        """
        return str(value)

    def process_result_value(self, value, dialect):
        if not value or value == "None":
            value = 0
        return Decimal(value)


class TransactionOperationString(TypeDecorator):
    impl = sqlalchemy.String

    def process_bind_param(self, value, dialect):
        """ Needs to return an object of the underlying impl
        """
        assert isinstance(value, Transaction.Operation)
        return str(value.value)

    def process_result_value(self, value, dialect):
        """ Load from the DB and turn into an enum """
        if not value:
            value = Transaction.Operation.NOOP
        return Transaction.Operation(value)


@enum.unique
class Market(enum.Enum):
    """
    Note that we don't use auto() because it's not in python3.5
    """

    BTCUSD = 1
    ETHUSD = 2
    BCHUSD = 3
    ZECUSD = 4
    LTCUSD = 5
    BTCETH = 6


@enum.unique
class Symbol(enum.Enum):
    """
    These aren't in wide use yet.
    TODO: Migrate to use these instead of strings.
    """

    BTC = 1
    ETH = 2
    BCH = 3
    USD = 4
    LTC = 5
    ZEC = 6


def is_fiat(symbol):
    return symbol == "USD"


class Transaction(yabc.Base):
    """
    Exchange-independent representation of a transaction (buy or sell).

    Each transaction is something that the user actually clicked a button to do. For example, trading BTC for ETH is one
    transaction.

    On Binance, a single trade can result in several events that affect taxes. Consider the BTCETH market, and a single SELL order executed there:
    - coins can be used to pay fees. this means the coins need to be subtracted from the pool.
    - BTC needs to be removed from the pool
    - ETH needs to be added (less fees)
    """

    @enum.unique
    class Operation(enum.Enum):
        NOOP = "Noop"
        BUY = "Buy"
        SELL = "Sell"
        GIFT_RECEIVED = "GiftReceived"
        GIFT_SENT = "GiftSent"
        MINING = "Mining"
        SPENDING = "Spending"
        # The following aren't stored in the DB, put typically only used in basis calculations as temporary values.
        TRADE_INPUT = "TradeInput"
        SPLIT = "Split"

    __tablename__ = "transaction"
    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
    asset_name = sqlalchemy.Column(sqlalchemy.String)  # Deprecated
    quantity = sqlalchemy.Column(PreciseDecimalString)  # Deprecated
    date = sqlalchemy.Column(sqlalchemy.DateTime)
    fees = sqlalchemy.Column(PreciseDecimalString)
    fee_symbol = sqlalchemy.Column(sqlalchemy.String)  # new in schema 8
    operation = sqlalchemy.Column(TransactionOperationString)

    quantity_traded = sqlalchemy.Column(PreciseDecimalString)  # column added
    symbol_traded = sqlalchemy.Column(sqlalchemy.String)  # column added

    quantity_received = sqlalchemy.Column(PreciseDecimalString)  # column added
    symbol_received = sqlalchemy.Column(sqlalchemy.String)  # column added

    source = sqlalchemy.Column(sqlalchemy.String)
    usd_subtotal = sqlalchemy.Column(PreciseDecimalString)  # deprecated
    user_id = sqlalchemy.Column(sqlalchemy.Integer, sqlalchemy.ForeignKey("user.id"))

    def __init__(
        self,
        operation: Operation,
        asset_name="BTC",  # deprecated
        date=None,
        fees=0,
        quantity=0,
        source=None,
        usd_subtotal=0,  # deprecated
        user_id="",
        symbol_traded="",
        symbol_received="",
        quantity_traded=0,
        quantity_received=0,
        fee_symbol="USD",
    ):
        for param in (quantity, fees, usd_subtotal):
            assert isinstance(param, (float, str, Decimal, int))
        self.quantity = Decimal(quantity)
        self.operation = operation
        self.date = date
        if date and isinstance(date, datetime.datetime):
            self.date = date.replace(tzinfo=None)
        self.usd_subtotal = Decimal(usd_subtotal)
        self.source = source
        self.asset_name = asset_name
        self.user_id = user_id
        self.fees = Decimal(fees)

        # The new syntax explicitly labels each leg of the trade.
        # Allow its use to overwrite values in columns for "asset_name" and "usd_subtotal"
        self.quantity_received = Decimal(quantity_received)
        self.quantity_traded = Decimal(quantity_traded)
        self.fee_symbol = fee_symbol
        self.symbol_received = symbol_received
        self.symbol_traded = symbol_traded

        if symbol_received or symbol_traded:
            # Both must be specified.
            # Quantity can't be used.
            # Try to populate the old fields.
            assert symbol_received and symbol_traded
            assert not quantity
            if self.is_simple_input():
                self.quantity = quantity_received
                self.asset_name = symbol_received
                self.usd_subtotal = quantity_traded
            else:
                # We're disposing of the asset.
                self.asset_name = symbol_traded
                self.usd_subtotal = quantity_received
                self.quantity = quantity_traded

        self.init_on_load()
        assert self.quantity_traded or self.quantity_received

    def is_coin_to_coin(self):
        if self.operation in {
            Operation.MINING,
            Operation.GIFT_SENT,
            Operation.GIFT_RECEIVED,
            Operation.SPLIT,
        }:
            return False
        if self.operation in {Operation.BUY, Operation.SELL}:
            return not is_fiat(self.symbol_traded) and not is_fiat(self.symbol_received)

    def needs_migrate_away_from_asset_name(self):
        return self.symbol_received == "" and self.symbol_traded == ""

    @sqlalchemy.orm.reconstructor
    def init_on_load(self):
        """
        Restore fields from self.asset_name and self.quantity to the more general columns.
        """
        if not self.needs_migrate_away_from_asset_name():
            # Do not modify the object further if we've already restored fields.
            return

        self.fee_symbol = (
            "USD"
        )  # Not possible to have others, until binance or other coin/coin markets are added.
        if self.is_simple_input():
            self.symbol_received = self.asset_name
            self.quantity_received = self.quantity
            self.symbol_traded = "USD"
            self.quantity_traded = self.usd_subtotal
        else:
            # Check the assumption that there are no SPLITs in the DB.
            assert self.operation in {
                Transaction.Operation.SPENDING,
                Transaction.Operation.GIFT_SENT,
                Transaction.Operation.SELL,
            }
            # We assume that these are SELLs in the sense that we are trading away BTC and receiving cash.
            self.symbol_traded = self.asset_name
            self.quantity_traded = self.quantity
            self.symbol_received = "USD"
            self.quantity_received = self.usd_subtotal

    def is_simple_input(self):
        return self.operation in {
            Operation.MINING,
            Operation.GIFT_RECEIVED,
            Operation.SPLIT,
            Operation.BUY,
            Operation.TRADE_INPUT,
        }

    def trade_has_input(self):
        """
        TODO: coin/coin trades make this more complicated, as SELLs can also be inputs.
        :return: True if this transaction is an input (like mining, a gift received, or a purchase)
        """
        if self.operation in {
            Operation.MINING,
            Operation.GIFT_RECEIVED,
            Operation.SPLIT,
            Operation.BUY,
            Operation.TRADE_INPUT,
        }:
            return True
        if self.Operation == Operation.SELL:
            return self.is_coin_to_coin()
        return False

    def is_taxable_output(self):
        return self.operation in {Operation.SPENDING, Operation.SELL}

    def __repr__(self):
        return "<TX {date} {operation} {rcvd} {rcvd_symbol} for {traded} {traded_symbol}, from exchange {source}. Fee {fee} {feecoin}>".format(
            date=self.date,
            traded=self.quantity_traded,
            traded_symbol=self.symbol_traded,
            rcvd=self.quantity_received,
            rcvd_symbol=self.symbol_received,
            operation=self.operation.name,
            source=self.source,
            fee=self.fees,
            feecoin=self.fee_symbol,
        )


Operation = Transaction.Operation



"""
Code for importing data from a report generated by the exchange Coinbase.

YABC is not affiliated with the company Coinbase or any of its products.
"""
import csv
import decimal

from dateutil import parser

from yabc import transaction
from yabc.formats import FORMAT_CLASSES
from yabc.formats import Format


def from_coinbase(f):
    """
    :param f: a file-like object with csv data.
    @return dictionaries with coinbase fields
    """
    f.seek(0)
    rawcsv = [i for i in csv.reader(f)]
    if len(rawcsv) < 5:
        raise RuntimeError("Invalid CSV file, not enough rows.")
    fieldnames = rawcsv[4]
    if not len(fieldnames) >= 2 or not fieldnames[-2].count("Coinbase") > 0:
        raise RuntimeError("Invalid coinbase file encountered")
    fieldnames[-1] = "Bitcoin Hash"
    fieldnames[-2] = "Coinbase ID"
    f.seek(0)
    transactions = [i for i in csv.DictReader(f, fieldnames)]
    # Previously, coinbase column 4 had timestamp.  Documents generated as
    # of April 2019 have it as the first column.
    # First meaningful row is 4; 3 has headers so we can check that.
    assert transactions[3]["Timestamp"] == "Timestamp"
    transactions = transactions[4:]
    for i in transactions:
        assert "Coinbase ID" in i
        i["Site"] = "Coinbase"
    USD_and_back = [
        i
        for i in transactions
        if (i["Transfer Total"] is not "" and i["Transfer Total"] is not None)
    ]
    return USD_and_back


def txs_from_coinbase(f):
    """
    :param f: a filelike object with CSV data
    :return: a list of transaction.Transaction
    """
    dicts = from_coinbase(f)
    return [FromCoinbaseJSON(i) for i in dicts]


class CoinbaseParser(Format):
    EXCHANGE_NAME = "Coinbase"

    def __init__(self, file_or_fname):
        self._file = file_or_fname
        self._reports = []
        if isinstance(file_or_fname, str):
            with open(file_or_fname) as f:
                self._reports = txs_from_coinbase(f)
        else:
            self._reports = txs_from_coinbase(file_or_fname)

    def __iter__(self):
        return self

    def __next__(self):
        if not self._reports:
            raise StopIteration
        return self._reports.pop(0)


def FromCoinbaseJSON(json):
    """
    Arguments:
        json (dict): a coinbase-style dictionary with the following fields:
            - 'Transfer Total': the total USD price, including fees.
            - 'Transfer Fee': the USD fee charged by the exchange.
            - 'Amount': the amount of bitcoin sold. (It's negative for sales.)
            - 'Currency': which cryptocurrency was involved.
            - 'Timestamp': 'hour:min:sec.millisecs' formatted timestamp.
    Returns: Transaction instance with important fields populated
    """
    operation = transaction.Transaction.Operation.BUY
    proceeds = json["Transfer Total"]
    fee = json["Transfer Fee"]
    asset_name = json["Currency"]
    quantity = decimal.Decimal(json["Amount"])
    if quantity < 0:
        operation = transaction.Transaction.Operation.SELL
        quantity = abs(quantity)
    timestamp_str = json["Timestamp"]
    return transaction.Transaction(
        asset_name=asset_name,
        operation=operation,
        quantity=quantity,
        date=parser.parse(timestamp_str),
        fees=fee,
        source="coinbase",
        usd_subtotal=proceeds,
    )


FORMAT_CLASSES.append(CoinbaseParser)
